package hopp

import "io"
import "fmt"
import "net"
import "sync"
import "sync/atomic"
import "git.tebibyte.media/sashakoshka/hopp/tape"
import "git.tebibyte.media/sashakoshka/go-util/sync"

// TODO investigate why 30 never reaches the server, causing it to wait for ever
// and never close the connection, causing the client to also wait forever

const closeMethod = 0xFFFF
const int64Max = int64((^uint64(0)) >> 1)
const defaultChunkSize = 0x1000

// Party represents a side of a connection.
type Party bool; const (
	ServerSide Party = false
	ClientSide Party = true
)

func (party Party) String() string {
	if party == ServerSide {
		return "server"
	} else {
		return "client"
	}
}

type a struct {
	sizeLimit    int64
	underlying   net.Conn
	party        Party
	transID      int64
	transLock    sync.RWMutex
	sendLock     sync.Mutex
	transMap     map[int64] *transA
	transChan    chan *transA
	done         chan struct { }
	err          error
}

// AdaptA returns a connection implementing METADAPT-A over a singular stream-
// oriented transport such as TCP or UNIX domain stream sockets.
func AdaptA(underlying net.Conn, party Party) Conn {
	conn := &a {
		sizeLimit:  defaultSizeLimit,
		underlying: underlying,
		party:      party,
		transMap:   make(map[int64] *transA),
		transChan:  make(chan *transA),
		done:       make(chan struct { }),
	}
	if party == ClientSide {
		conn.transID =  1
	} else {
		conn.transID = -1
	}
	go conn.receive()
	return conn
}

func (this *a) Close() error {
	close(this.done)
	return nil
}

func (this *a) LocalAddr() net.Addr {
	return this.underlying.LocalAddr()
}

func (this *a) RemoteAddr() net.Addr {
	return this.underlying.RemoteAddr()
}

func (this *a) OpenTrans() (Trans, error) {
	this.transLock.Lock()
	defer this.transLock.Unlock()
	if this.transID == int64Max {
		return nil, fmt.Errorf("could not open transaction: %w", ErrIntegerOverflow)
	}
	id := this.transID
	trans := &transA {
		parent:   this,
		id:       id,
		incoming: usync.NewGate[incomingMessage](),
	}
	this.transMap[id] = trans
	if this.party == ClientSide {
		this.transID ++
	} else {
		this.transID --
	}
	return trans, nil
}

func (this *a) AcceptTrans() (Trans, error) {
	eof := fmt.Errorf("could not accept transaction: %w", io.EOF)
	select {
	case trans := <- this.transChan:
		if trans == nil {
			return nil, eof
		}
		return trans, nil
	case <- this.done:
		return nil, eof
	}
}

func (this *a) SetSizeLimit(limit int64) {
	this.sizeLimit = limit
}

func (this *a) unlistTransactionSafe(id int64) {
	this.transLock.Lock()
	defer this.transLock.Unlock()
	delete(this.transMap, id)
}

func (this *a) sendMessageSafe(trans int64, method uint16, data []byte) error {
	this.sendLock.Lock()
	defer this.sendLock.Unlock()
	return encodeMessageA(this.underlying, this.sizeLimit, trans, method, data)
}

func (this *a) receive() {
	defer func() {
		this.underlying.Close()
		close(this.transChan)
		this.transLock.Lock()
		defer this.transLock.Unlock()
		for _, trans := range this.transMap {
			trans.closeDontUnlist()
		}
		clear(this.transMap)
		this.underlying.Close()
	}()

	// receive MMBs in a loop and forward them to transactions until shit
	// starts closing
	for {
		transID, method, chunked, payload, err := decodeMessageA(this.underlying, this.sizeLimit)
		if err != nil {
			this.err = fmt.Errorf("could not receive message: %w", err)
			return
		}

		err = this.multiplexMMB(transID, method, chunked, payload)
		if err != nil {
			this.err = fmt.Errorf("could not receive message: %w", err)
			return
		}
	}
}

func (this *a) multiplexMMB(transID int64, method uint16, chunked bool, payload []byte) error {
	if transID == 0 { return ErrMessageMalformed }

	trans, err := func() (*transA, error) {
		this.transLock.Lock()
		defer this.transLock.Unlock()

		trans, ok := this.transMap[transID]
		if !ok {
			// check if this is a superfluous close message and just
			// do nothing if so
			if method == closeMethod {
				return nil, nil
			}
			
			// it is forbidden for the other party to initiate a transaction
			// with an ID from this party
			if this.party == partyFromTransID(transID) {
				return nil, ErrMessageMalformed
			}
			trans = &transA {
				parent:   this,
				id:       transID,
				incoming: usync.NewGate[incomingMessage](),
			}
			this.transMap[transID] = trans
			this.transChan <- trans
		}
		return trans, nil
	}()
	if err != nil { return err }

	if trans == nil {
		return nil
	}

	if method == closeMethod {
		return trans.Close()
	} else {
		trans.incoming.Send(incomingMessage {
			method:  method,
			chunked: chunked,
			payload: payload,
		})
	}
	return nil
}

// most methods in transA don't need to be goroutine safe except those marked
// as such
type transA struct {
	parent        *a
	id            int64
	incoming      usync.Gate[incomingMessage]
	currentReader io.Reader
	currentWriter io.Closer
	writeBuffer   []byte
	closed        atomic.Bool
}

func (this *transA) Close() error {
	// MUST be goroutine safe
	err := this.closeDontUnlist()
	this.parent.unlistTransactionSafe(this.ID())
	return err
}

func (this *transA) closeDontUnlist() (err error) {
	// MUST be goroutine safe
	this.incoming.Close()
	if !this.closed.Load() {
		err = this.Send(closeMethod, nil)
	}
	this.closed.Store(true)
	return err
}

func (this *transA) ID() int64 {
	return this.id
}

func (this *transA) Send(method uint16, data []byte) error {
	return this.parent.sendMessageSafe(this.id, method, data)
}

func (this *transA) SendWriter(method uint16) (io.WriteCloser, error) {
	// close previous writer if necessary
	if this.currentWriter != nil {
		this.currentWriter.Close()
		this.currentWriter = nil
	}

	// create new writer
	writer := &writerA {
		parent: this,
		// there is only ever one writer at a time, so they can all
		// share a buffer
		buffer: this.writeBuffer[:0],
		method: method,
		chunkSize: defaultChunkSize,
		open: true,
	}
	this.currentWriter = writer
	return writer, nil
}

func (this *transA) Receive() (method uint16, data []byte, err error) {
	method, reader, err := this.ReceiveReader()
	if err != nil { return 0, nil, err }
	data, err = io.ReadAll(reader)
	if err != nil { return 0, nil, err }
	return method, data, nil
}

func (this *transA) ReceiveReader() (uint16, io.Reader, error) {
	// if the transaction has been closed, return an io.EOF
	if this.closed.Load() {
		return 0, nil, io.EOF
	}

	// drain previous reader if necessary
	if this.currentReader != nil {
		io.Copy(io.Discard, this.currentReader)
	}

	// create new reader
	reader := &readerA {
		parent: this,
	}
	method, err := reader.pull()
	if err != nil { return 0, nil, err}
	this.currentReader = reader
	return method, reader, nil
}

type readerA struct {
	parent   *transA
	leftover []byte
	eof      bool
}

// pull pulls the next MMB in this message from the transaction.
func (this *readerA) pull() (uint16, error) {
	// if the previous message ended the chain, return an io.EOF
	if this.eof {
		return 0, io.EOF
	}

	// get an MMB from the transaction we are a part of
	receive := this.parent.incoming.Receive()
	if receive != nil {
		if message, ok := <- receive; ok {
			if message.method != closeMethod {
				this.leftover = append(this.leftover, message.payload...)
				if !message.chunked {
					this.eof = true
				}
				return message.method, nil
			}
		}
	}

	// close and return error on failure
	this.eof = true
	this.parent.Close()
	if this.parent.parent.err == nil {
		return 0, fmt.Errorf("could not receive message: %w", io.EOF)
	} else {
		return 0, this.parent.parent.err
	}
}

func (this *readerA) Read(buffer []byte) (int, error) {
	if len(this.leftover) == 0 {
		if this.eof { return 0, io.EOF }
		this.pull()
	}

	copied := copy(buffer, this.leftover)
	this.leftover = this.leftover[copied:]
	return copied, nil
}

type writerA struct {
	parent    *transA
	buffer    []byte
	method    uint16
	chunkSize int64
	open      bool
}

func (this *writerA) Write(data []byte) (n int, err error) {
	if !this.open { return 0, io.EOF }
	toSend := data
	for len(toSend) > 0 {
		nn, err := this.writeOne(toSend)
		n += nn
		toSend = toSend[nn:]
		if err != nil { return n, err }
	}
	return n, nil
}

func (this *writerA) Close() error {
	this.open = false
	return nil
}

func (this *writerA) writeOne(data []byte) (n int, err error) {
	data = data[:min(len(data), int(this.chunkSize))]
	
	// if there is more room, append to the buffer and exit
	if int64(len(this.buffer) + len(data)) <= this.chunkSize {
		this.buffer = append(this.buffer, data...)
		n = len(data)
		// if have a full chunk, flush
		if int64(len(this.buffer)) == this.chunkSize {
			err = this.flush()
			if err != nil { return n, err }
		}
		return n, nil
	}

	// if not, flush and store as much as we can in the buffer
	err = this.flush()
	if err != nil { return n, err }
	this.buffer = append(this.buffer, data...)
	return n, nil
}

func (this *writerA) flush() error {
	return this.parent.parent.sendMessageSafe(this.parent.id, this.method, this.buffer)
}

type incomingMessage struct {
	method  uint16
	chunked bool
	payload []byte
}

func encodeMessageA(
	writer    io.Writer,
	sizeLimit int64,
	trans     int64,
	method    uint16,
	data      []byte,
) error {
	if int64(len(data)) > sizeLimit {
		return ErrPayloadTooLarge
	}
	buffer := make([]byte, 18 + len(data))
	tape.EncodeI64(buffer[:8], trans)
	tape.EncodeI16(buffer[8:10], method)
	tape.EncodeI64(buffer[10:18], uint64(len(data)))
	copy(buffer[18:], data)
	_, err := writer.Write(buffer)
	return err
}

func decodeMessageA(
	reader    io.Reader,
	sizeLimit int64,
) (
	transID       int64,
	method        uint16,
	chunked       bool,
	payloadBuffer []byte,
	err           error,
) {
	headerBuffer := [18]byte { }
	_, err = io.ReadFull(reader, headerBuffer[:])
	if err != nil { return 0, 0, false, nil, err }
	transID, err = tape.DecodeI64[int64](headerBuffer[:8])
	if err != nil { return 0, 0, false, nil, err }
	method, err = tape.DecodeI16[uint16](headerBuffer[8:10])
	if err != nil { return 0, 0, false, nil, err }
	size, err := tape.DecodeI64[uint64](headerBuffer[10:18])
	if err != nil { return 0, 0, false, nil, err }
	chunked, size = splitCCBSize(size)
	if size > uint64(sizeLimit) {
		return 0, 0, false, nil, ErrPayloadTooLarge
	}
	payloadBuffer = make([]byte, int(size))
	_, err = io.ReadFull(reader, payloadBuffer)
	if err != nil { return 0, 0, false, nil, err }
	return transID, method, chunked, payloadBuffer, nil
}

func partyFromTransID(id int64) Party {
	return id > 0
}

func splitCCBSize(size uint64) (bool, uint64) {
	return size >> 63 > 1, size & 0x7FFFFFFFFFFFFFFF
}
